/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/

#ifndef COMM_H
#define COMM_H

#include <inttypes.h>
#include <avr/io.h>
#include "global.h"

#ifdef GPSLOG
#include "gps_parse.h"
#endif

#ifdef TRIGGERFRAME_INPUT
#include "triggerframe.h"
#endif

#include "spi.h"
#include "ringbuf.h"
#include "slcan.h"

/** MTsendRTvar() runtime bytes */
#define MT_RTVAR_COUNT (56)
#define MT_RTVAR_COUNTMORE (75)
// round up to even number: 75 => 76   76=>76
#define MT_RTVAR_COUNTMORE_X ((MT_RTVAR_COUNTMORE + 1) & 0xFE) 

enum MegatunePageIdx
{
    MT_PAGE_IDX_CONFIG = 0,     // config
    MT_PAGE_IDX_HWMAP,          // hw map
    MT_PAGE_IDX_BOOST,          // boost structure
    MT_PAGE_IDX_CONFIGEXT,      // config_ext
    MT_PAGE_IDX_TEMPERATURE,    // ve.mat_enrich table
    MT_PAGE_IDX_A340E,          // at structure
    MT_PAGE_IDX_LAUNCHRPM,	// launchrpm table
    MT_PAGE_IDX_CONFIG_SHARED,

/* following tables all have 2 axis in them (usually RPM and KPA).
* Table size must be specified (8x8, not 8x8+8+8),
* get_bins_pointer() decides which pages share bins => ranges are sent automagically.
* Could be changed, so only LAMBDAVIEW is included and everything after ...IGNADV
*/
    PAGE_IDX_FIRST_RPMKPA,
    // labda size
    MT_PAGE_IDX_LAMBDA = PAGE_IDX_FIRST_RPMKPA, // labmda (lambda_size default 8x8)
    // LAMBDAVIEW is same data as LAMBDA, but 1/x style for retarded tuningprograms
    MT_PAGE_IDX_LAMBDAVIEW,         // labmda (lambda_size default 8x8)
    // ve size
    MT_PAGE_IDX_VE,             	// ve (ve_size default 16x14 in my_make)
    MT_PAGE_IDX_CHARGE,
    MT_PAGE_IDX_MAPCORR,            // map_corr (map_corr_size default 8x8)
    MT_PAGE_IDX_MATRET,

// following tables all share rpmkpa with ign_adv table:
// ign size
    MT_PAGE_IDX_IGNADV,             // ign_adv (adv_size default 12x12)
    MT_PAGE_IDX_BOOST_REF,          // boost ref (adv_size default 12x12)
    MT_PAGE_IDX_1JZVVTI,            // vvti_adv (adv_size default 12x12)
    MT_PAGE_IDX_KNOCK,              // knock ref, was 8*8 (+8 +8)
// please UNDERSTAND page_has_rpmkpa before choosing index for a new table (likely must be placed before VE)
// UPDATE this when adding new table:
    MT_PAGE_IDX_LAST = MT_PAGE_IDX_KNOCK
};
uint8_t MEGATUNE_PAGE_SIZE(uint8_t page);
uint8_t* MEGATUNE_PAGE_POINTER(uint8_t page);
#define MT_PAGE_CRC_PACKET_LEN ((3*(MT_PAGE_IDX_LAST+1))+4)

// config12 bit0 was hijacked for ETC !
#define ALTERNATIVE_RT_DATA 0
#define alternative_rtdata()	(config.knock_noise_scale & _BV(ALTERNATIVE_RT_DATA))

#define MT_PAGE_IDX_DUMMY (0xFF)
#define MT_PAGE_SIZE_DUMMY (0)

// only pages that have idx greater or equal MT_PAGE_IDX_VE clone the rpm and kpa axis of the VE table
#define page_has_rpmkpa(pageidx) (pageidx >= PAGE_IDX_FIRST_RPMKPA && pageidx != MT_PAGE_IDX_KNOCK)

#define MT_CODEVER_LEN (26)

struct privch_t {
	uint8_t a;
	uint8_t b;
};

struct bins_pointer_t {
	uint8_t *p_kpa_range, *p_rpm_range;
	uint8_t kpa_size, rpm_size;
};

// note: this is accidentally 8 bytes, and 8 entries, but gcc uses 16 bit operation <<3 in a cycle
// also, cmd_queue could be on 256 byte boundary
struct cmd_queue_t {
  uint8_t (*func)(uint8_t i);	// null is special, means send bytes from (SRAM) private[]
  uint8_t len;    // number of bytes to send
  uint8_t count;  // number of bytes already sent
  union {
    uint8_t * private;
    uint8_t (*privatefunc)(uint8_t i);
  };

  union {
  	struct privch_t privch;
   	// direct access of privw (eg. using privch) is endian-dependent!!!
   	// Do NOT access privw using privch.
   	// Only use privch in case when not accessed as privw
   	// http://www.avrfreaks.net/index.php?module=FreaksFiles&func=viewFile&id=271&showinfo=1
   	// use LO() and HI() macros
	uint16_t privw;
  };
#ifdef SIPR
	uint8_t seq;   // sequence number used in sipr
#endif
};

#define LO(x)	(x & 0xFF)
#define HI(x)	((x >> 8) & 0xFF)

#define MT_LAMBDA_OFFSET 73
#define LAMBDA_INT_OFFSET 200


// We could enumerate instead, but ansi-C promotes enumerations to
// integers (16 bit), no need for that!
// shared variable with TRIGGERFRAME_INPUT data state without overlaping , need reinit on mode changing
#define CMD    0
#define OFFSET 1
#define VALUE  2

#define TRIGGERFRAME_UNESCAPE 0xE0 //comm_state == 0x0E triggerframe input process, next character need unescape

//com.status defines:
// #define MODBUS_ON_COMM 2 // my_make... undef: no MODBUS, 1: MODBUS ON UART0,  2: MODBUS ON UART1.
#define PRIMTRIG_NOT_CLOSED 0// shows if an headless primary trigger log in progress
#define MEGATUNE_WRITE_DISABLED 1//disable crc-less megatune write command after calling the stepping to triggerframe mode the first time
#define MODBUS_RTVAR_LEN  MT_RTVAR_COUNTMORE_X+5 // old command +6: +4 for preamble modbusaddr,cmd='D',offset,len +2 for CRC. +5 is enough now
#ifdef MODBUS_SLAVE
#define UART1_BUFF_LEN  MODBUS_RTVAR_LEN    //UART1 rx buffer length,  (longest modbus rtu command, previously 21)
#else // not MODBUS_SLAVE
#define UART1_BUFF_LEN  21  // maybe even shorter would be sufficient ? 9..10 bytes ? Longest query from MASTER is apparently only 6+2 bytes
#endif // not MODBUS_SLAVE
// #define MT_RTVAR_FIRST_SEI 52  // first sei at 52, so 0..51 might be sent even from interrupt context (if cycles is not a problem, eg. at 1500 RPM)
struct com_t {
    uint8_t comm_state;
    uint16_t commcrc; // outgoing CRC
	uint16_t incoming_CRC; // but in MENU_MODBUS this is wallclock time of last character received (and com.commcrc used for incoming CRC calc)
	uint8_t status;
    uint8_t comm1_relay_allowed;//if not zero relays com1 to com0 (decrements every 100ms)
    uint8_t modbus_rx_state; // only used for modbus, could be merged with comm_state: 0=this frame not interesting; 1-255: frame might be interesting
    uint8_t modbus_rx_cnt; // only used for modbus, could be merged with status  // 0-7 receiving "command 05" byte 0-7; 128-135: receiving "command 04" byte 0-7; 192-199: receiving "command 02" (m7065 Read discrete inputs)
#ifdef GPSLOG //allocated size is 24byte
    struct gps_t gps;
#endif
#ifdef TRIGGERFRAME_INPUT //allocated size is RINGBUFLEN + 4
    struct triggerframe_t triggerframe;
#endif
    uint8_t atomic_helper; // store 2nd byte when "sending" 1st byte of 16 bit value. Added atomic_helper_more[] array right after if another source needs this (now only SD needs it, comm functions store value in cmdqueue)
    // DO NOT INSERT ANYTHING HERE, it would be corrupted by atomic_helper
#if MODBUS_ON_COMM == 2
    uint8_t atomic_helper_more[1]; // keep it RIGHT after atomic_helper ! +N number of bytes for atomic_helper as more sources need it
    // uint8_t uart1_fill_idx; // if nonzero, uart1_tx_buff[ --uart1_fill_idx ] has YET TO BE FILLED by userspace. Currently just 'D' flags that 'D' command response needs to be generated from userspace
    uint8_t uart1_tx_idx; //initial length : if nonzero, send  uart1_tx_buff[--uart1_tx_idx] next...
    uint8_t uart1_tx_startingpoint; // startingpoint+1 actually, so 0 => 1 (nonzero)
    uint8_t uart1_tx_buff[UART1_BUFF_LEN];
#endif
};

void userspace_fill_uart1_tx_buff(void); // link only succeeds if MODBUS_ON_COMM == 2

#ifdef ONE_WIRE
//status states:
#define ONE_WIRE_LISTENING 0    //default state, wait for incoming 1-wire data
#define ONE_WIRE_CRC_FAIL 1     //1-wire data received, crc check failed
#define ONE_WIRE_CRC_OK 2       //1-wire data received, crc check passed
#define ONE_WIRE_FOUND_CORRECT 3//found everything ok: (1-wire crc and data) or (adc limit)
#define ONE_WIRE_FOUND_INCORRECT 4 //crc ok, but code incorrect, need it for testing ibutton and show it
// #define ONE_WIRE_ADC_CHECK 5    //waiting for adc measure ready

// SRAM reading revealed that for some reason after-boot reading is higher than expected (atmega128a starts quicker and AREF is low?)
// higher than 0x40 (~1.25V ?) (eg. 0x46 or 0x42 even after minute of discharge manually with DVM to 0.22V)
// after 10 seconds reading was 0x6B
// Aty found 0x70 is reached too quickly (sometimes 3 sec)
// 0x5E .. 0x68 seems good (in 1.2.13 changed 0x5E=> 68)
// new AVR-s seem to not discharge, needed minutes (even charge a bit when ign on ?), so upped to 6C (cell 2016-05 Also changed HW from 4.7uF*1.5MOhm to 1uF*1MOhm)
#define ONE_WIRE_ADC_LIMIT ( 0x6C )
struct one_wire_t {
    uint8_t status;//see the defines above
    uint8_t tx_cnt;//number of next byte to send
    uint8_t rx_cnt;//number of bytes received
    uint8_t rx_crc;//8bit crc for the incoming 1byte family code and 6 byte serial code
    uint8_t msB;//most significant byte of the serial code   TODO: kill this byte: compare when available and adjust status
    uint8_t ssB;//second significant byte of the serial code TODO: kill this byte: compare when available and adjust status
};
extern struct one_wire_t one_wire;
#endif //ONE_WIRE

extern struct com_t com;
extern uint8_t cmd_queue_head, cmd_queue_tail;
#define cmd_queue_empty() (cmd_queue_head == cmd_queue_tail)

// see TABLES_TO_DUMP* before changing this
// CMD_QUEUE_SIZE must be a power of 2 with the cheap define, no such limitation with the smarter comm_buffer_NOT_full() inline function
#define CMD_QUEUE_SIZE 8
#define comm_buffer_NOT_full() (cmd_queue_head != ((cmd_queue_tail + 1) % CMD_QUEUE_SIZE))
/* smarter than the define: it can accept CMD_QUEUE_SIZE which is not power of 2. But +2 bytes flash.
static inline uint8_t comm_buffer_NOT_full( void )
{
	uint8_t h1 = cmd_queue_head;

	if( h1 ) h1--; // nonzero, no overflow
	else h1 = CMD_QUEUE_SIZE-1;
	return h1 - cmd_queue_tail; // if( cmd_queue_tail == h1 ) return 1; else return 0;
}
*/

extern struct cmd_queue_t cmd_queue[CMD_QUEUE_SIZE];

// WBO2 does not use comm_TX_plainbytes any longer...
// KNOCK_DETAILED_LOG neither, comm_TX_plainbytes() uses mmcbuffer
// #define comm_TX_plain_SIZE (10)
// EXTERN uint8_t comm_TX_plainbytes[comm_TX_plain_SIZE] SECTION(".mmcbuffer"); // ideally after the mmc buffer (so this gets corrupted by stack first)

#ifdef STACKPOINTER_DEBUG
// not needed when STACKPOINTER_DEBUG not used, but we enable now... Shouldn't cause problem at the top of heap
EXTERN uint8_t profiling[3] SECTION(".lcdsect");
#endif

static uint16_t get_stackpointer(void){ return (SPH<<8) + SPL; }
/* static inline void update_stackpointer_lowest(void) ... painting is much better
*{
*	uint16_t stackp = get_stackpointer();
*	if((stackp < stackpointer_lowest) || (stackpointer_lowest== 0)) stackpointer_lowest = stackp;
*} */
/** paints 1,2,3.. values from stackpointer-1 downto 0x1000 to help debug interrupt stack usage */
static inline void paint_marginzone(void)
{
	uint16_t stackp = get_stackpointer();
	uint8_t b=0;
	while(--stackp >= 0x1000) *(uint8_t*)stackp = ++b; // 1,2,3, ... as we going down
}



/** returns 12 bit mcp3208 data. Can be called from userspace or interrupt.
* inputch & 7 : truncated to 3 bits */
static inline uint16_t get_mcp3208_data12(uint8_t inputch)
{
#ifndef MCP3208_USERSPACE
    uint16_t ret16;
    uint8_t sreg_bkup;
#ifndef INPUT_MUX_USED		
    inputch &= 7;
#else
	inputch &= (MCP3208_CHANNELS-1);
#endif	
    sreg_bkup = SREG;
    cli();
    ret16 = mcp3208_data[ inputch ];
    SREG = sreg_bkup;
    // if( ret16 >= 4096 ) ret16 = mcp3208_data[ inputch ]; // with another attempt (and FF-ing the high-byte before writing the L,H from interrupt), the cli() could be omitted (small gain)
    return ret16;
#else
#ifndef INPUT_MUX_USED		
    inputch &= 7;
#else
	inputch &= (MCP3208_CHANNELS-1);
#endif
	return mcp3208_data[ inputch ];
#endif	// MCP3208_USERSPACE
	
}



void initUART(void);
void initUART1(void);
void comm(uint8_t data);

uint8_t sendStuffedSimple(uint8_t cnt) SECTION(".selfpushsect");
uint8_t send_page_len_crc_stuffed (uint8_t cnt) SECTION(".selfpushsect");
uint8_t SendPage_data_stuffed (uint8_t cnt) SECTION(".selfpushsect2");
uint8_t SendPage_stuffed( uint8_t i ) SECTION(".selfpushsect");
/* sendpage error codes*/
#define SENDPAGE_ERROR_NO_PAGEONID 1

uint8_t MTsendRTvar_triggerlog(uint8_t i) SECTION(".selfpushsect2");
uint8_t MTsendRTvar(uint8_t i, uint8_t caller); //i must be monotonous increasing!!, if caller==0 store ATOMIC_HELPER in cmdqueue, otherwise in com.atomic_helper global var
uint8_t MTtestComm(uint8_t i);
uint8_t MTcodeVer(uint8_t i);

void change_page(uint8_t page);
void get_bins_pointer(uint8_t page, struct bins_pointer_t *bins_pointer);
uint8_t send_page(uint8_t i);
void store_page(uint8_t addr, uint8_t value);
uint8_t sendConfigVar(uint8_t i);
uint8_t allSensorRaw(uint8_t i);
uint8_t ConfigDump(uint8_t cnt);
uint8_t TableDump(uint8_t cnt);
uint8_t TriggerlogPrimtrig(uint8_t cnt) SECTION(".selfpushsect");
//TriggerlogAnytrig type defines:
#define TRIGGERFRAME_IGNDEACT 8
// cylinder specific ignact from 80..87 (88,89 reserved), and igndeact from 90..97  (98,99 reserved):
#define TRIGGERFRAME_IGNACT_CYL0 80
#define TRIGGERFRAME_IGNDEACT_CYL0 90
#define TRIGGERFRAME_PTRIG_AT0 9
uint8_t TriggerlogAnytrig( uint8_t i ) SECTION(".selfpushsect");
uint8_t TriggerlogAnytrigPar( uint8_t i ) SECTION(".selfpushsect");
uint8_t pushfunc(uint8_t (*f)(uint8_t i), uint8_t len);
void pushfuncpar2(uint8_t (*f)(uint8_t i), uint8_t len, void * par, uint8_t p2);
void pushfuncpar2W(uint8_t (*f)(uint8_t i), uint8_t len, void * par, uint16_t p2);
void pushfuncpar(uint8_t (*f)(uint8_t i), uint8_t len, void *);
void pushfuncTableDump(const uint8_t *p, uint8_t entries);
uint8_t pushfuncsipr(uint8_t (*f)(uint8_t i), uint8_t len, uint8_t seq);

uint8_t srl_direct(uint8_t);
uint8_t srl_direct_hex( uint8_t i );
uint8_t srl_generic(uint8_t);
uint8_t srl_generic_atomic16_crc(uint8_t addr);
uint8_t srl_generic_E(uint8_t addr) SECTION(".selfpushsect2");
uint8_t srl_generic_P(uint8_t addr);
uint8_t srl_generic_P_upper(uint8_t addr);
uint8_t sendSignature(uint8_t i);

static __inline__ void ringbuf_push_HDLC_END_commcrc_ffff(void)
{
	ringbuf_push(&comm_sndbuf, HDLC_END);
	com.commcrc = 0xffff;
}

void status_com1(void);
#ifdef ONE_WIRE
void read_1wire(void);
void check_1wire_adc(void);
void one_wire_adc_DIScharge(void);
#endif //ONE_WIRE

/* user defined baudrate */
static inline void change_baudrate( uint8_t newbaud ){
	UBRR0H = 0;
	UBRR0L = newbaud;
}

/* baudrate change pending flag. Ignored when 0,
 * otherwise it contains the new baudrate divisor */
extern uint8_t new_baudrate;

extern const uint8_t PROGMEM TABLES_TO_DUMP1[];
extern const uint8_t PROGMEM TABLES_TO_DUMP2[];

// Use for debugging
void put_char(uint8_t byte);

// #define MODBUS_DEBUG (config.cranking_thres > 0)
// #define MODBUS_DEBUG (0)
// or send lotsof junk for FF, only TIMOUT signal ("T" when FE):
#define MODBUS_DEBUG (config.cranking_thres == 0xFF)
#define MODBUS_DEBUG_SIGNALTIMEOUT (config.cranking_thres >= 0xFE)
#define MODBUS_TIMEOUT (((uint16_t)config.cranking_thres)+3)

#ifdef MODBUS_SEND_BYTES_ATBOOT
void command_buffer(void);
#endif

/// used  in sipr.c
uint8_t comm_getLambdaView(uint8_t val);
uint8_t comm_setLambdaView(uint8_t val);
#endif
